"""
Example solution for the afternoon Exercise of Day 1.

written by Niko Wilbert for the G-Node Python Summer School 2009

This module contains all elements for an abstract graph representation.
The concrete solution for the exercise is in the city_graph.py file. 
"""

from heapq import heappop, heappush


class Node(object):
    """Base class for nodes.
    
    To add additional data attributes to nodes you can derive subclasses.
    """
    
    def __init__(self):
        """Initialize the node."""
        self._out_edges = []  # outgoing edges
        self._in_edges = []  # incoming edges
        
    @property
    def out_edges(self):
        return self._out_edges
    

class Edge(object):
    """Edge class.
    
    This class also registers itself with the affected nodes, so it sits
    between Node and Graph in the hierarchy.
    """
    
    def __init__(self, head, tail):
        """Initialize the edge.
        
        head, tail -- The end and starting node of this edge.
        """
        self._head = head
        if head is not None:
            head._in_edges.append(self)
        self._tail = tail
        if tail is not None:
            tail._out_edges.append(self)
        
    def set_head(self, head):
        """Set the head and register this in the nodes as well.
        
        The head can also be none, then the edge is unregistered.
        """
        if self._head is not None:
            self._head._in_edges.remove(self)
        self._head = head
        if head is not None:
            self._head._in_edges.append(self)
            
    def set_tail(self, tail):
        """Set the tail and register this in the nodes as well."""
        if self._head is not None:
            self._tail._out_edges.remove(self)
        self._tail = tail
        if tail is not None:
            self._tail._out_edges.append(self)
            
    def clear_nodes(self):
        """Clear the node references."""
        self.set_tail(None)
        self.set_head(None)
        
    head = property(lambda self: self._head, set_head)
    tail = property(lambda self: self._tail, set_tail)


class GraphException(Exception):
    """Base Exception for Graph."""
    pass


class NoPathGraphException(GraphException):
    """Base Exception for Graph."""
    pass


class Graph(object):
    """Class to represent a complete graph with nodes and edges.
    
    Note that this class does not support nodes which do not belong to any
    edge (internally only edges are stored).
    """
    
    def __init__(self, edges=None):
        self._edges = set()
        if edges:
            for edge in edges:
                self.add_edge(edge)
                
    @property
    def edges(self):
        """Return list of the edges in this graph."""
        return self._edges
    
    @property
    def nodes(self):
        """Return set of nodes in this graph."""
        nodes = set()
        for edge in self._edges:
            nodes.add(edge.head)
            nodes.add(edge.tail)
        return nodes
        
    def add_edge(self, edge):
        """Add an edge to the graph."""
        self._edges.add(edge)
        
    def remove_edge(self, edge):
        """Remove an edge from the graph."""
        edge.clear_nodes()
        self._edges.remove(edge)
        
    def shortest_path(self, start_node, end_node, weight_func=lambda edge: 1):
        """Return the shortest path and its overall weight.
        
        weight -- Maps an edge to a weight value, the default
            maps all edges to 1.
        """
        # store the shortest path to all nodes,
        # the value tuples contain the path length and the previous node 
        shortest_paths = {start_node: (0, None)}
        # we use this list as a priority cue with heapq
        # the tuples contain the overall path length and the edge itself
        edge_heap = []
        for edge in start_node.out_edges:
            heappush(edge_heap, (weight_func(edge), edge))
        while edge_heap:
            path_weight, edge = heappop(edge_heap)
            if ((edge.head not in shortest_paths) or
                (shortest_paths[edge.head][0] > path_weight)):
                shortest_paths[edge.head] = (path_weight, edge)
                # if we already visited this node then there may be edge
                # duplicates in the heap, but this is no ptoblem because
                # the newer edges are guaranteed to come first
                for out_edge in edge.head.out_edges: 
                    heappush(edge_heap, (path_weight + weight_func(out_edge),
                                         out_edge))
        if end_node not in shortest_paths:
            err = ("The is no connection from node %s" % str(start_node) +
                   " to node %s." % str(end_node))
            raise NoPathGraphException(err)
        # assemble the shortest path from end to start
        path_weight = shortest_paths[end_node][0]
        path_edges = [shortest_paths[end_node][1]]
        current_node = path_edges[-1].tail
        while current_node is not start_node:
            path_edges.append(shortest_paths[current_node][1])
            current_node = path_edges[-1].tail
        return path_edges[::-1], path_weight
        
